"""
Provides the Dataset class
"""

from __future__ import annotations

from typing import Optional, Literal
import warnings
import numpy as np
import dask.array as da
import scipy.sparse as sparse

from .typing import Modality, Axis, DataTensor
from .extras.prior import Prior
from .numbafied import extract_d_values

# Optional dependencies
try:
    from anndata import AnnData
except ImportError:
    AnnData = None
try:
    from mudata import MuData
except ImportError:
    MuData = None

class Dataset:
    def __init__(
        self,
        *,
        dataset: dict[Modality, DataTensor],
        structure: dict[Modality, tuple[Axis]],
        batch_axes: Optional[set[Axis]] = None
    ):
        # Type declarations of attributes
        self.dataset: dict[Modality, DataTensor]
        self.structure: dict[Modality, tuple[Axis]]

        self.batch_axes: set[Axis]
        self.presences: dict[Axis, dict[Modality, Optional[int]]]
        self.presences_batchless: dict[Axis, dict[Modality, Optional[int]]]

        self.full_sizes: dict[Modality, np.ndarray]
        self.left_sizes: dict[Modality, np.ndarray]
        self.right_sizes: dict[Modality, np.ndarray]
        self.Ks: dict[str, int]

        self.axis_sizes: dict[Axis, int]

        # Type declarations of cached attributes
        # calculated via sklearn api calls
        self.prior: dict[Axis, Prior]
        self.mean: dict[Axis, float]
        self.gram_matrices: dict[Axis, np.ndarray]
        self.es: dict[Axis, np.ndarray]
        self.evecs: dict[Axis, np.ndarray]
        self.evals: dict[Axis, np.ndarray]
        self.precision_matrices: dict[Axis, np.ndarray]

        # Input handling for default arguments
        if batch_axes is None:
            self.batch_axes = set({""})
        else:
            self.batch_axes = batch_axes

        # Set attributes
        self.dataset = dataset
        self.structure = structure
        self.gram_matrices = {}
        self.prior = {}
        self.es = {}
        self.evecs = {}
        self.evals = {}
        self.precision_matrices = {}

        # Calculate helper attributes
        self.recompute_helpers()

        # Make cached attributes for covariance subdivision
        self.cache()

        # Initialize variable that stores random state
        self.random_state: Optional[int] = None

    def memory_usage(self) -> tuple[int, int, int]:
        """
        Returns the memory usage of the dataset
        """
        dataset_bytes: int = 0
        aux_bytes: int = 0
        precision_bytes: int = 0

        # Keep track of dataset size
        if self.old_dataset is self.dataset:
            for tensor in self.dataset.values():
                dataset_bytes += array_bytes(tensor)
        else:
            # If parts of the dataset has been cached,
            # report the cached dataset as the uncached bits
            # are views, i.e. they do not take up any memory
            for tensor in self.old_dataset.values():
                dataset_bytes += array_bytes(tensor)

        # Add all intermediat results
        for gram_matrix in self.gram_matrices.values():
            aux_bytes += array_bytes(gram_matrix)
        for evals in self.evals.values():
            aux_bytes += array_bytes(evals)
        for evecs in self.evecs.values():
            aux_bytes += array_bytes(evecs)

        # Add output
        for precision_matrix in self.precision_matrices.values():
            precision_bytes += array_bytes(precision_matrix)

        # Potentially add cached gram matrices
        if self.gram_matrices is not self.old_gram_matrices:
            for gram_matrix in self.old_gram_matrices.values():
                aux_bytes += array_bytes(gram_matrix)

        return dataset_bytes, aux_bytes, precision_bytes
    
    def print_memory_usage(self) -> None:
        """
        Prints the memory usage of the dataset
        """
        dataset_bytes, aux_bytes, precision_bytes = self.memory_usage()
        print(f"Dataset: {dataset_bytes / 1e9:.2f} GB")
        print(f"Auxiliary: {aux_bytes / 1e9:.2f} GB")
        print(f"Output: {precision_bytes / 1e9:.2f} GB")

    def cache(self) -> None:
        self.old_dataset = self.dataset
        self.old_structure = self.structure
        self.old_all_axes = self.all_axes
        self.old_axis_sizes = self.axis_sizes
        self.old_gram_matrices = self.gram_matrices

    def uncache(self) -> None:
        self.dataset = self.old_dataset
        self.structure = self.old_structure
        self.gram_matrices = self.old_gram_matrices
        self.recompute_helpers()


    def recompute_helpers(self) -> None:
        self._calculate_axes()
        self._calculate_presences()
        self._calculate_dimensions()
        self._calculate_axis_sizes()

    def _calculate_axes(self) -> None:
        """
        Gets a set of every axis in the dataset, i.e.
        [("A", "B"), ("A", "B")]
        => {"A", "B", "C"}
        """
        self.all_axes: set[str] = {
            axis
            for axes in self.structure.values()
            for axis in axes
        } - self.batch_axes

    def _calculate_presences(self) -> None:
        """
        Returns a dictionary keyed by axis names,
        indicating which datasets the axis is present in

        We assume each axis appears at most once per dataset
        """

        self.presences: dict[Axis, dict[Modality, Optional[int]]] = {}
        self.presences_batchless: dict[Axis, dict[Modality, Optional[int]]] = {}

        for axis in self.all_axes:
            self.presences[axis] = {}
            self.presences_batchless[axis] = {}
            for modality, axes in self.structure.items():
                to_add: list[int] = [
                    idx
                    for idx, ax
                    in enumerate(axes)
                    if ax == axis
                ]

                if len(to_add) == 0:
                    self.presences[axis][modality] = None
                    self.presences_batchless[axis][modality] = None
                elif len(to_add) == 1:
                    presence: int = to_add[0]
                    batches_before_presence: int = len(
                        [
                            ax
                            for ax in axes[:presence]
                            if ax in self.batch_axes
                        ]
                    )
                    self.presences[axis][modality] = presence
                    self.presences_batchless[axis][modality] = presence - batches_before_presence
                else:
                    raise Exception("GmGM does not allow repeated axes in a dataset.")
                
    def _calculate_dimensions(self) -> None:
        """
        Get d_\forall, d_{<\ell}, and d_{>\ell}, and K for each modality,
        as defined in the paper.
        """
        self.full_sizes: dict[Modality, np.ndarray] = {}
        self.left_sizes: dict[Modality, np.ndarray] = {}
        self.right_sizes: dict[Modality, np.ndarray] = {}

        for modality, tensor in self.dataset.items():
            full_size, left_size, right_size = extract_d_values(
                tensor.shape
            )
            self.full_sizes[modality] = full_size
            self.left_sizes[modality] = left_size
            self.right_sizes[modality] = right_size

        # Get the number of dimensions for each tensor
        self.Ks: dict[str, int] = {
            modality: len(
                set(self.structure[modality]) - self.batch_axes
            )
            for modality in self.dataset.keys()
        }

    def _calculate_axis_sizes(self) -> None:
        """
        Calculates the size of each axis
        """
        self.axis_sizes: dict[Axis, int] = {}
        for axis in self.all_axes:
            self.axis_sizes[axis] = [
                tensor.shape[idx]
                for modality, tensor in self.dataset.items()
                if (idx := self.presences[axis][modality]) is not None
            ][0]

    def __repr__(self) -> str:

        # Print structure of each dataset
        dataset_str: str = "\n".join(
            f"\t{modality}: {axes}"
            for modality, axes in self.structure.items()
        )

        # Print calculation status info
        prior_str: dict[Axis, str] = {
            axis: "Prior: " + (self.prior[axis].name if axis in self.prior else "None")
            for axis in self.all_axes
        }
        gram_str: dict[Axis, str] = {
            axis: "Gram: " + ("Calculated" if axis in self.gram_matrices else "Not calculated")
            for axis in self.all_axes
        }
        eig_str: dict[Axis, str] = {
            axis: "Eig: " + ("Calculated" if axis in self.evecs else "Not calculated")
            for axis in self.all_axes
        }

        # Print the size of each axis
        size_str: str = "\n".join(
            f"\t{axis}: {size}\n\t\t"
            + "\n\t\t".join([x[axis] for x in [prior_str, gram_str, eig_str]])
            for axis, size in self.axis_sizes.items()
        )

        return f"Dataset(\n{dataset_str}\n)\nAxes(\n{size_str}\n)"
    
    @classmethod
    def from_AnnData(
        cls,
        data: AnnData,
        use_highly_variable: bool = True
    ) -> Dataset:
        if AnnData is None:
            raise ImportError("Please install AnnData to use this method.")
        matrix = data.X
        if use_highly_variable and 'highly_variable' in data.var.keys():
            matrix = data[:, data.var.highly_variable].X

        dataset = Dataset(
            dataset={
                "AnnData" : matrix#da.from_array(matrix)
            },
            structure={
                "AnnData": ("obs", "var")
            }
        )
        dataset.base = data
        dataset.base_use_highly_variable = use_highly_variable
        return dataset
    
    def to_AnnData(
        self,
        key_added: str = "gmgm",
        use_abs_of_graph: bool = True
    ) -> AnnData:
        if AnnData is None:
            raise ImportError("Please install AnnData to use this method.")
        if self.base is None:
            raise ValueError("This dataset was not created from an AnnData object.")

        _add_graph_to_anndata(
            self.base,
            self.base_use_highly_variable,
            "obs",
            self.base.shape[0],
            "obs",
            key_added,
            self.precision_matrices["obs"],
            self.base.obs.highly_variable \
                if "highly_variable" in self.base.obs \
                else None,
            use_abs_of_graph,
            self.random_state
        )
        _add_graph_to_anndata(
            self.base,
            self.base_use_highly_variable,
            "var",
            self.base.shape[0],
            "var",
            key_added,
            self.precision_matrices["var"],
            self.base.var.highly_variable \
                if "highly_variable" in self.base.obs \
                else None,
            use_abs_of_graph,
            self.random_state
        )

        return self.base
    
    @classmethod
    def from_MuData(
        cls,
        data: MuData,
        use_highly_variable: bool = True
    ) -> Dataset:
        if MuData is None:
            raise ImportError("Please install MuData to use this method.")
        
        # Is MuData concatenating along the features axis?
        if data.axis == 0:
            matrices: dict[Modality, np.ndarray] = {}
            for modality in data.mod:
                matrix = data[modality].X
                if use_highly_variable and 'highly_variable' in data[modality].var.keys():
                    matrix = data[modality][:, data[modality].var.highly_variable].X
                matrices[modality] = matrix

            dataset = Dataset(
                dataset=matrices,
                structure={
                    modality: ("obs", f"{modality}-var")
                    for modality in data.mod
                }
            )
            dataset.base = data
            dataset.base_use_highly_variable = use_highly_variable
            return dataset
        # Is MuData concatenating along the samples axis?
        elif data.axis == 1:
            raise NotImplementedError("Concatenating along the samples axis is not yet supported.")
        # Is MuData concatenating along both axes?
        elif data.axis == -1:
            raise NotImplementedError("Concatenating along both axes is not yet supported.")
        else:
            raise ValueError(f"Invalid concatenation axis {data.axis}.")
    
    def to_MuData(
        self,
        key_added: str = "gmgm",
        use_abs_of_graph: bool = True
    ) -> MuData:
        if MuData is None:
            raise ImportError("Please install MuData to use this method.")
        if self.base is None:
            raise ValueError("This dataset was not created from a MuData object.")
        
        # Is MuData concatenating along the features axis?
        if self.base.axis == 0:
            # First add the observations axis to the MuData object
            _add_graph_to_anndata(
                self.base,
                self.base_use_highly_variable,
                "obs",
                self.base.shape[0],
                "obs",
                key_added,
                self.precision_matrices["obs"],
                self.base.obs.highly_variable \
                    if "highly_variable" in self.base.obs \
                    else None,
                use_abs_of_graph,
                self.random_state
            )
            
            # Then, for each modality, add the features axis to the MuData object
            for modality in self.base.mod:
                _add_graph_to_anndata(
                    self.base[modality],
                    self.base_use_highly_variable,
                    f"{modality}-var",
                    self.base[modality].shape[1],
                    "var",
                    key_added,
                    self.precision_matrices[f"{modality}-var"],
                    self.base[modality].var.highly_variable \
                        if "highly_variable" in self.base[modality].var \
                        else None,
                    use_abs_of_graph,
                    self.random_state
                )
        # Is MuData concatenating along the samples axis?
        elif self.base.axis == 1:
            raise NotImplementedError("Concatenating along the samples axis is not yet supported.")
        # Is MuData concatenating along both axes?
        elif self.base.axis == -1:
            raise NotImplementedError("Concatenating along both axes is not yet supported.")
        else:
            raise ValueError(f"Invalid concatenation axis {self.base.axis}.")
        
        return self.base
    
    def modalities_with_axis(self, axis: Axis) -> list[int, Modality]:
        """
        Returns a list of modalities that have the given axis,
        paired with the location of the axis in the modality
        """
        return [
            (list(axes).index(axis), modality)
            for modality, axes in self.structure.items()
            if axis in axes
        ]
    
def array_bytes(
    arr: np.ndarray | sparse.sparray
) -> float:
    """
    Returns the number of bytes required to store
    a numpy array or sparse matrix
    """
    if isinstance(arr, sparse.spmatrix):
        raise ValueError(
            "Please use sparse arrays instead of the to-be-deprecated"
            + " sparse matrix API."
        )
    
    # Input handling for various sparse array types
    if isinstance(arr, sparse.csr_array) | isinstance(arr, sparse.csc_array):
        return (
            arr.data.nbytes
            + arr.indices.nbytes
            + arr.indptr.nbytes
        )
    elif isinstance(arr, sparse.coo_array):
        return (
            arr.data.nbytes
            + arr.row.nbytes
            + arr.col.nbytes
        )
    else:
        return arr.nbytes
    
def _add_graph_to_anndata(
    adata: AnnData | MuData,
    use_highly_variable: bool,
    axis: Axis,
    axis_size: int,
    obs_or_var: Literal["obs", "var"],
    key_added: str,
    graph: np.ndarray | sparse.sparray | sparse.spmatrix,
    highly_variable_indices: Optional[np.ndarray] = None,
    use_abs_of_graph: bool = True,
    random_state: Optional[int] = None
) -> AnnData | MuData:
    
    if use_abs_of_graph:
        graph = abs(graph)
    
    adata.uns[f'{axis}_neighbors_{key_added}'] = {
        'connectivities_key': f'{axis}_{key_added}_connectivities',
        'distances_key': f'{axis}_{key_added}_connectivities',
    }
    adata.uns[f'{axis}_neighbors_{key_added}']['params'] = {
        'method': 'gmgm',
    }
    if random_state is not None:
        adata.uns[f'{axis}_neighbors_{key_added}']['params']['random_state'] = random_state

    dictp: dict
    if obs_or_var == "obs":
        dictp = adata.obsp
    elif obs_or_var == "var":
        dictp = adata.varp
    else:
        raise ValueError(f"Invalid `obs_or_var`: {obs_or_var}")

    if use_highly_variable and highly_variable_indices is None:
        warnings.warn(
            "`use_highly_variable` was set to True but no highly variable indices were provided"
            + f" for axis '{axis}'.  Thus we will use all indices for this axis."
        )

    if use_highly_variable and highly_variable_indices is not None:
        # Create empty full matrix
        out_graph = sparse.coo_array((axis_size, axis_size))
        
        # Fill with highly variable connectivities
        loc = np.where(highly_variable_indices)[0]
        graph = graph.tocoo()
        out_graph.data = graph.data
        out_graph.row = loc[graph.row]
        out_graph.col = loc[graph.col]
        
        # Convert to an efficient matrix representation
        dictp[f'{axis}_{key_added}_connectivities'] = \
            out_graph.tocsr()
    else:
        dictp[f'{axis}_{key_added}_connectivities'] = graph.tocsr()

    return adata


from .dataset import Dataset
from .GmGM import GmGM
from .core.core import direct_svd, calculate_eigenvalues, calculate_eigenvectors
from .core.preprocessing import center, clr_prost, create_gram_matrices
from .core.presparse_methods import recompose_sparse_precisions

__all__ = [
    "Dataset",
    "GmGM",
    "direct_svd",
    "calculate_eigenvalues",
    "calculate_eigenvectors",
    "center",
    "clr_prost",
    "create_gram_matrices",
    "recompose_sparse_precisions",
]


"""
Some bits of code were originally written in Fortran,
however I have now ported them to Python+Numba as:
1) Pure python code is easier to distribute
2) Numba actually turned out to be faster!

I think the reason why is that I do not know much about
compilers so I did not choose the best settings to enable
parallelism, but Numba makes it quite easy.
"""

import numba as nb
import numpy as np
from .typing import Axis, Modality

class _project_inv_kron_sum:
    @staticmethod
    @nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
    def two_axis(
        x: np.ndarray,
        y: np.ndarray
    ) -> tuple[np.ndarray]:
        
        k_ratio: float = 1./2.
        x_out: np.ndarray = np.zeros(x.shape[0])
        y_out: np.ndarray = np.zeros(y.shape[0])

        for i in nb.prange(x.shape[0]):
            for j in nb.prange(y.shape[0]):
                cur_val: float = 1 / (x[i]+y[j])
                x_out[i] += cur_val
                y_out[j] += cur_val

        # Normalize
        x_out /= x.shape[0]
        y_out /= y.shape[0]

        # Offset diagonal
        x_out -= k_ratio * np.sum(x_out) / x.shape[0]
        y_out -= k_ratio * np.sum(y_out) / y.shape[0]

        return x_out, y_out

    @staticmethod
    @nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
    def three_axis(
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray
    ) -> tuple[np.ndarray]:
        
        k_ratio: float = 2./3.
        x_out: np.ndarray = np.zeros(x.shape[0])
        y_out: np.ndarray = np.zeros(y.shape[0])
        z_out: np.ndarray = np.zeros(z.shape[0])

        for i in nb.prange(x.shape[0]):
            for j in nb.prange(y.shape[0]):
                for k in nb.prange(z.shape[0]):
                    cur_val: float = 1 / (x[i]+y[j]+z[k])
                    x_out[i] += cur_val
                    y_out[j] += cur_val
                    z_out[j] += cur_val

        # Normalize
        x_out /= x.shape[0]
        y_out /= y.shape[0]
        z_out /= z.shape[0]

        # Offset diagonal
        x_out -= k_ratio * np.sum(x_out) / x.shape[0]
        y_out -= k_ratio * np.sum(y_out) / y.shape[0]
        z_out -= k_ratio * np.sum(z_out) / z.shape[0]

        return x_out, y_out, z_out
    
    @nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
    def four_axis(
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        w: np.ndarray
    ) -> tuple[np.ndarray]:
        
        k_ratio: float = 3./4.
        x_out: np.ndarray = np.zeros(x.shape[0])
        y_out: np.ndarray = np.zeros(y.shape[0])
        z_out: np.ndarray = np.zeros(z.shape[0])
        w_out: np.ndarray = np.zeros(w.shape[0])

        for i in nb.prange(x.shape[0]):
            for j in nb.prange(y.shape[0]):
                for k in nb.prange(z.shape[0]):
                    for l in nb.prange(w.shape[0]):
                        cur_val: float = 1 / (x[i]+y[j]+z[k]+w[l])
                        x_out[i] += cur_val
                        y_out[j] += cur_val
                        z_out[j] += cur_val
                        w_out[j] += cur_val

        # Normalize
        x_out /= x.shape[0]
        y_out /= y.shape[0]
        z_out /= z.shape[0]
        w_out /= w.shape[0]

        # Offset diagonal
        x_out -= k_ratio * np.sum(x_out) / x.shape[0]
        y_out -= k_ratio * np.sum(y_out) / y.shape[0]
        z_out -= k_ratio * np.sum(z_out) / z.shape[0]
        w_out -= k_ratio * np.sum(w_out) / w.shape[0]

        return x_out, y_out, z_out, w_out
    
class _sum_log_sum:
    @staticmethod
    @nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
    def two_axis(
        x: np.ndarray,
        y: np.ndarray,
        simplify_size: int = 20
    ) -> float:
        
        sl: float = 0
        intermediate: float = 1
        
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(y.shape[0]):
                sm: float = x[i] + y[j]
                intermediate *= sm

                if (i+j) % simplify_size == 0:
                    sl += np.log(intermediate)
                    intermediate = 1.0

        sl += np.log(intermediate)

        return sl
    
    @staticmethod
    @nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
    def three_axis(
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        simplify_size: int = 20
    ) -> float:
        
        sl: float = 0
        intermediate: float = 1
        
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(y.shape[0]):
                for k in nb.prange(z.shape[0]):
                    sm: float = x[i] + y[j] + z[k]
                    intermediate *= sm

                    if (i+j+k) % simplify_size == 0:
                        sl += np.log(intermediate)
                        intermediate = 1.0

        sl += np.log(intermediate)

        return sl
    
    @staticmethod
    @nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
    def four_axis(
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        w: np.ndarray,
        simplify_size: int = 20
    ) -> float:
        
        sl: float = 0
        intermediate: float = 1
        
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(y.shape[0]):
                for k in nb.prange(z.shape[0]):
                    for l in nb.prange(w.shape[0]):
                        sm: float = x[i] + y[j] + z[k] + w[l]
                        intermediate *= sm

                        if (i+j+k+l) % simplify_size == 0:
                            sl += np.log(intermediate)
                            intermediate = 1.0

        sl += np.log(intermediate)

        return sl

_proj_iks: dict[int, callable] = {
    2: _project_inv_kron_sum.two_axis,
    3: _project_inv_kron_sum.three_axis,
    4: _project_inv_kron_sum.four_axis,
}
_sls: dict[int, callable] = {
    2: _sum_log_sum.two_axis,
    3: _sum_log_sum.three_axis,
    4: _sum_log_sum.four_axis,
}
          

def project_inv_kron_sum(
    evals: dict[Axis, np.ndarray],
    structure: dict[Modality, tuple[Axis]],
    modality: Modality,
    batch_axes: set[Axis],
    K: int
):
    args = [
        evals[axis]
        for axis in structure[modality]
        if axis not in batch_axes
    ]
    
    if K not in _proj_iks:
        raise ValueError(
            "Only 2-4 way tensors are supported!"
            + f"... What are you doing with {K} axes?"
        )
    return _proj_iks[K](*args)

def sum_log_sum(
        *args: list[np.array]
    ) -> float:
        """
        Computes:
            the sum
            of the log
            of the determinant
            of the kronecker sum
            of the input matrices 
        """
        K = len(args)
        if K not in _sls:
            raise ValueError(
                "Only 2-4 way tensors are supported!"
                + f"... What are you doing with {K} axes?"
            )
        return _sls[K](*args)

def extract_d_values(shape):
    d_all = np.prod(shape)
    d_lefts = np.cumprod((1, *shape))[:-1]
    d_rights = np.cumprod((1, *shape[::-1]))[::-1][1:]

    return d_all, d_lefts, d_rights


"""
Provides the Prior class
"""

from __future__ import annotations

from typing import Literal
import numpy as np

class Prior:
    def __init__(self) -> None:
        pass

class Wishart(Prior):
    name = "Wishart"
    def __init__(
        self,
        *,
        degrees_of_freedom: int,
        scale_matrix: np.ndarray
    ):
        self.degrees_of_freedom = degrees_of_freedom
        self.scale_matrix = scale_matrix

        p: int = self.scale_matrix.shape[0]

        assert self.scale_matrix.shape == (p, p), \
            "Scale matrix must be square"
        
        assert degrees_of_freedom > p - 1, \
            "Degrees of freedom must be greater than p - 1"
        
        self.subtract_from_gram = -np.linalg.inv(self.scale_matrix)

    def process_gram(
        self,
        gram: np.ndarray
    ) -> np.ndarray:
        # Eta = -1/2 * prior^-1
        # Drop the 1/2 as we ignore it everywhere
        # Subtract from Gram as in Theorem 3
        return gram - self.subtract_from_gram
    
    def process_gradient(
        self,
        evals: np.ndarray
    ) -> np.ndarray:
        # h = -1/2 * precmat^-1
        # drop the 1/2 as we ignore it everywhere
        return -1/evals


"""
This file contains the methods that are meant to be used
before eigendecomposition/singular value decomposition

Specifically, this includes centering, grammifying, 
and nonparanormal-skeptic-related methods.
"""

from __future__ import annotations
from typing import Optional

from ..dataset import Dataset
from ..typing import Axis
import scipy.stats as stats
import scipy.sparse as sparse
import numpy as np

import warnings

def center(
    X: Dataset
) -> Dataset:
    """
    'centers' the dataset by subtracting each dataset's mean value
    Note: not row-mean or column-mean, but overall mean.

    Modifies X in place
    """

    for modality in X.dataset.keys():
        if sparse.issparse(X.dataset[modality]):
            warnings.warn(
                "Dataset was sparse but `avg-overall` centering will desparsify it,"
                + " consider using a different centering method."
                + "  Converting to dense array..."
            )
            X.dataset[modality] = X.dataset[modality].toarray()
        X.dataset[modality] -= X.dataset[modality].mean()

    return X

def clr_prost(
    X: Dataset,
) -> Dataset:
    """
    Takes the modified clr as described in
    "A zero inflated log-normal model for inference of sparse microbial association networks"
    by Prost et al. (2021)

    Don't use on already-logp1 transformed data

    Involves taking mean of all axes other than the first

    Modifies X in place
    """

    for modality in X.dataset.keys():
        dataset = X.dataset[modality]
        if not sparse.issparse(dataset):
            axes = tuple(range(1, len(dataset.shape)))
            dataset[dataset != 0] = np.log(dataset[dataset != 0])
            dataset -= np.sum(dataset, axis=axes, keepdims=True) / np.count_nonzero(dataset, axis=axes, keepdims=True)
        else:
            dataset = dataset.tocoo()
            dataset.data = np.log(dataset.data)
            dataset = dataset.tocsr()
            nonzero = dataset.nonzero()
            dataset[nonzero] -= (dataset.sum(axis=1).A1 / dataset.getnnz(axis=1))[nonzero[0]]

        X.dataset[modality] = dataset

def create_gram_matrices(
    X: Dataset,
    batch_size: Optional[int] = None,
    use_nonparanormal_skeptic: bool = False
) -> Dataset:
    """
    Modifies X in place
    """
    # Clear old gram matrices
    X.gram_matrices = {}

    # Calculate effective gram matrices
    num_samples: dict[Axis, int] = {}
    
    for modality, tensor in X.dataset.items():
        for idx, axis in enumerate(X.structure[modality]):
            if axis in X.batch_axes:
                continue

            if sparse.issparse(tensor):
                warnings.warn(
                    "Dataset was sparse but `create_gram_matrices` will desparsify it."
                    + "  This may be avoidable if you set the `n_comps` when calling `GmGM`."
                    + "  Converting to dense array..."
                )
                tensor = tensor.toarray()
                X.dataset[modality] = tensor

            matricized: np.ndarray = np.reshape(
                np.moveaxis(tensor, idx, 0),
                (tensor.shape[idx], -1),
            )

            if axis in X.gram_matrices:
                X.gram_matrices[axis] += _grammify_core(
                    matricized,
                    batch_size,
                    use_nonparanormal_skeptic
                )
                num_samples[axis] += X.full_sizes[modality] // X.axis_sizes[axis]
            else:
                X.gram_matrices[axis] = _grammify_core(
                    matricized,
                    batch_size,
                    use_nonparanormal_skeptic
                )
                num_samples[axis] = X.full_sizes[modality] // X.axis_sizes[axis]

    # Divide gram matrices by effective number of modalities of samples seen
    for axis in X.gram_matrices.keys():
        X.gram_matrices[axis] /= num_samples[axis]

    return X

def _grammify_core(
    matricized: np.ndarray,
    batch_size: Optional[int] = None,
    use_nonparanormal_skeptic: bool = False
) -> np.ndarray:
    """
    TODO: Check if this still works
    """
    
    output: np.ndarray
    if batch_size is None:
        batch_size = matricized.shape[0]

    if use_nonparanormal_skeptic:
        for idx in range(0, matricized.shape[0] + batch_size, batch_size):
            increase = min(batch_size, matricized.shape[0] - idx)
            if idx <= 0:
                break

            matricized[idx:idx+increase] = stats.rankdata(matricized[idx:idx+increase], axis=0)
            matricized[idx:idx+increase] -= matricized[idx:idx+increase].mean(axis=0, keepdims=True)

    output = matricized @ matricized.T
    if use_nonparanormal_skeptic:
        output = np.sin(np.pi/6 * output)
    
    return output



"""
This file contains the core methods of GmGM,
specifically Theorems 1 and 2 of the ArXiv paperß
"""

from __future__ import annotations
from typing import Optional

from ..dataset import Dataset
from ..typing import Axis, Modality
from ..numbafied import project_inv_kron_sum, sum_log_sum
import numpy as np
import dask.array as da
import scipy.sparse as sparse

import warnings

# TODO: Figure out what to do with these
from ..extras.regularizers import Regularizer
from ..extras.prior import Prior

def direct_left_eigenvectors(
    X: Dataset,
    n_comps: Optional[int] = None,
    n_power_iter: int = 4,
    n_over_samples: int = 100,
    random_state: Optional[int] = None
) -> Dataset:
    """
    No assumptions on `X` but works best when `n_comps` is specified
    """

    if n_comps is None:
        raise NotImplementedError("Need to specify `n_comps` for now")
    
    for axis in X.all_axes:
        if axis in X.batch_axes:
            continue
        to_concat: list[da.Array] = []
        for location, modality in X.modalities_with_axis(axis):
            tensor = X.dataset[modality]
            if sparse.issparse(tensor):
                warnings.warn(
                    "Dataset was sparse but `direct_left_eigenvectors` will desparsify it."
                    + "  This will change eventually."
                )
                tensor = tensor.toarray()
            matricized: da.array = da.from_array(np.reshape(
                np.moveaxis(tensor, location, 0),
                (tensor.shape[location], -1),
            ))
            to_concat.append(matricized)
        full_matricized: da.array = da.concatenate(to_concat, axis=1)
        V_1, Lambda, _ = da.linalg.svd_compressed(
            full_matricized,
            k=n_comps,
            compute=True,
            n_power_iter=4,
            n_oversamples=100,
            seed=random_state
        )
        Lambda = (Lambda**2).compute()
        V_1 = V_1.compute()
        X.evecs[axis] = V_1
        X.es[axis] = Lambda

    return X




def direct_svd(
    X: Dataset,
    n_comps: Optional[int] = None,
    n_power_iter: int = 4,
    n_over_samples: int = 100,
    random_state: Optional[int] = None
) -> Dataset:
    """
    Assumes `X` is a single matrix

    Split into four cases:
    (sparse, truncated)
    (sparse, not truncated)
    (dense, truncated)
    (dense, not truncated)
    """
    if len(X.dataset) != 1:
        raise ValueError("Dataset must be a single matrix")
    
    dataset = X.dataset[list(X.dataset.keys())[0]]
    if n_comps is not None:
        if sparse.issparse(dataset):
            V_1, Lambda, V_2 = sparse.linalg.svds(
                dataset,
                k=n_comps,
                random_state=random_state
            )
            Lambda = Lambda**2
            V_2 = V_2.T
        else:
            if not isinstance(dataset, da.Array):
                dataset = da.from_array(dataset)
            V_1, Lambda, V_2 = da.linalg.svd_compressed(
                dataset,
                k=n_comps,
                compute=True,
                n_power_iter=4,
                n_oversamples=100,
                seed=random_state
            )
            Lambda = (Lambda**2).compute()
            V_1 = V_1.compute()
            V_2 = V_2.T.compute()
    else:
        if sparse.issparse(dataset):
            warnings.warn(
                "Dataset was sparse but need density to compute all svds,"
                + " consider using the `n_comps` parameter."
                + "  Converting to dense array..."
            )
            dataset = dataset.toarray()
        if not isinstance(dataset, da.Array):
            dataset = da.from_array(dataset)
        V_1, Lambda, V_2 = da.linalg.svd(dataset)
        Lambda = (Lambda**2).compute()
        V_1 = V_1.compute()
        V_2 = V_2.T.compute()

    first_axis, second_axis = list(X.structure.values())[0]

    X.evecs[first_axis] = V_1
    X.evecs[second_axis] = V_2
    X.es[first_axis] = Lambda
    X.es[second_axis] = Lambda
    return X

def calculate_eigenvectors(
    X: Dataset,
    verbose: bool = False,
    n_comps: Optional[int] = None,
    random_state: Optional[int] = None,
    **params
) -> Dataset:
    # Initialize the gram matrices
    grams: dict[Axis, np.ndarray] = {}
    for axis in X.all_axes:
        grams[axis] = X.gram_matrices[axis]

    # Update the gram matrices that have priors
    for axis, prior in X.prior.items():
        grams[axis] = prior.process_gram(
            X.gram_matrices[axis]
        )

    # Next, calculate the eigenvalues and eigenvectors
    es: dict[Axis, np.ndarray] = {}
    evecs: dict[Axis, np.ndarray] = {}

    for axis, gram_matrix in grams.items():
        if verbose:
            print(f"Calculating eigenvectors for {axis=}")
        if n_comps is not None:
            if not isinstance(gram_matrix, da.Array):
                gram_matrix = da.from_array(gram_matrix)
            _, s, eigenvectors = da.linalg.svd_compressed(gram_matrix, seed=random_state, k=n_comps, **params)
            eigenvalues = s**2
            eigenvectors = eigenvectors.compute().T
            eigenvalues = eigenvalues.compute()
        else:
            if isinstance(gram_matrix, da.Array):
                gram_matrix = gram_matrix.compute()
            eigenvalues, eigenvectors = np.linalg.eigh(gram_matrix)

        es[axis] = eigenvalues
        evecs[axis] = eigenvectors

    # Convert eigenvectors to fortran order
    # This will make large multiplications faster, it seems!
    evecs = {
        axis: np.asfortranarray(evec)
        for axis, evec in evecs.items()
    }

    """
    When the trace term is large, we expect the eigenvalues to be small, as naive
    Gram matrix inversion would have the eigenvalues be the inverse of the trace term.
    This puts us close to the boundary of the positive definite cone, which is why
    it performs poorly - we have to tread to lightly and end up "converging" before
    we actually reach the optimum.

    Downscaling the trace term to be around 1 will mean the reciprocal also stays around 1,
    which keeps us away from the boundary.
    """
    es = {
        axis: es[axis] / np.linalg.norm(es[axis])
        for axis in X.all_axes
    }

    # Store the eigenvalues and eigenvectors
    X.es = es
    X.evecs = evecs

    return X

def calculate_eigenvalues(
    X: Dataset,
    *,
    max_small_steps: int = 5,
    max_line_search_steps: int = 20,
    lr_init: float = 1.0,
    max_iter: int = 1000,
    tol: float = 1e-3,
    regularizer: Optional[Regularizer] = None,
    force_posdef: bool = True,
    verbose: bool = False,
    verbose_every: int = 100,
    always_regularize: bool = False,
    check_overstep_each_iter: bool = False,
) -> Dataset:

    # Initialize looping variables
    num_small_steps: int = 0
    lr_t: float = lr_init
    prev_err: float = np.inf
    regularizing: bool = always_regularize and not regularizer is None

    # Eigenvalues of the mle, keyed by axis
    X.evals: dict[Axis, np.ndarray] = {
        axis: np.ones_like(e)
        for axis, e in X.es.items()
    }

    # Changes in eigenvalues in current iteration
    # keyed by axis
    diffs: dict[Axis, np.ndarray] = {
        axis: np.zeros_like(e)
        for axis, e in X.es.items()
    }

    # Converge to eigenvalue MLE
    for i in range(max_iter):

        # Compute MLE gradient
        for axis, locations in X.presences_batchless.items():
            assert axis not in X.batch_axes, \
                "Batch axes should not be in presences_batchless"
            diffs[axis] = X.es[axis].copy()
            for modality, ell in locations.items():
                if ell is None:
                    continue
                # Note to self: often it seems project_inv_kron_sum
                # can be removed, because it doesn't affect much...
                # Presumably the log determinant is much smaller in effect
                # than the trace...
                # Could be worth investigating how often this is the case
                diffs[axis] -= project_inv_kron_sum(
                    X.evals,
                    X.structure,
                    modality,
                    X.batch_axes,
                    X.Ks[modality],
                )[ell]

                for axis, prior in X.prior.items():
                    diffs[axis] += prior.process_gradient(
                        X.evals[axis]
                    )

        # Add regularization if necessary
        # Only activates after converging to the MLE
        if regularizing:
            regs: dict[str, np.ndarray] = regularizer.grad(
                X.evals,
                X.evecs,
            )
            for axis in X.all_axes:
                diffs[axis] += regs[axis]

        # Backtracking line search
        line_search_gave_up: bool = False
        lr_t: float = lr_init
        for line_step in range(max_line_search_steps):
            # Decrease step size each time
            # (`line_step` starts at 0, i.e. no decrease)
            step_lr: float = lr_t / 10**line_step
            
            for axis in X.all_axes:
                X.evals[axis] -= step_lr * diffs[axis]

            # Since all tuplets of eigenvalues
            # get summed together within each dataset,
            # the minimum final eigenvalue is 
            # the sum of minimum axis eigenvalues
            if not force_posdef:
                minimum_diag: float = min(
                    sum(
                        X.evals[axis].min()
                        for axis in axes
                        if axis not in X.batch_axes
                    )
                    for axes in X.structure.values()
                )
            # Or we can enforce individual posdefness!
            else:
                minimum_diag = min(
                min(
                    X.evals[axis].min()
                    for axis in axes
                    if axis not in X.batch_axes
                )
                for axes in X.structure.values()
                )

            # If the minimum eigenvalue is less than zero
            # have we left the positive definite space we desire
            # to stay in, so we will have to backtrack
            if minimum_diag <= 1e-8:
                for axis in X.all_axes:
                    X.evals[axis] += step_lr * diffs[axis]
                continue
            

            # Check if error has gotten worse
            if check_overstep_each_iter:
                log_err: float = _get_log_err(
                    X.evals,
                    X.structure
                )
                trace_err: float = _get_trace_err(
                    X.es,
                    X.evals,
                    X.all_axes
                )
                reg_err: float = _get_reg_err(
                    X.evals,
                    X.evecs,
                    regularizing,
                    regularizer
                )
                err: float = log_err + trace_err + reg_err
                if err > prev_err:
                    for axis in X.all_axes:
                        X.evals[axis] += step_lr * diffs[axis]
                    continue
            
            # If it got here, we have a good step size
            break
        else:
            # Did not find a good step size
            if verbose:
                print(f"@{i}: {prev_err} - Line Search Gave Up!")
            line_search_gave_up = True
            num_small_steps = max_line_search_steps + 1

        # Calculate the error
        if not check_overstep_each_iter:
            log_err: float = _get_log_err(
                X.evals,
                X.structure
            )
            trace_err: float = _get_trace_err(
                X.es,
                X.evals,
                X.all_axes
            )
            reg_err: float = _get_reg_err(
                X.evals,
                X.evecs,
                regularizing,
                regularizer
            )
            err: float = log_err + trace_err + reg_err

        # Calculate the change in error and
        # whether or not we can consider
        # ourselves to be converged
        err_diff: float = np.abs(prev_err - err)
        prev_err: float = err
        if err_diff/np.abs(err) < tol or line_search_gave_up:
            num_small_steps += 1
            if num_small_steps >= max_small_steps:
                if verbose:
                    print(f"Converged! (@{i}: {err})")
                if not regularizing and regularizer is not None:
                    if verbose:
                        print("Regularizing!")
                    tol /= 10
                    regularizing = True
                    num_small_steps = 0
                else:
                    break
        else:
            num_small_steps = 0

        if verbose:
            if i % verbose_every == 0:
                print(f"@{i}: {err} ({log_err} + {trace_err} + {reg_err}) ∆{err_diff / np.abs(err)}")
    else:
        # This triggers if we don't break out of the loop
        if verbose:
            print("Did not converge!")

    return X
    
def _get_log_err(
    evals: dict[Axis, np.ndarray],
    structure: dict[Modality, tuple[Axis]],
) -> float:
    """
    Calculates log-determinant portion of the error
    """
    log_err: float = 0
    for _, axes in structure.items():
        log_err -= sum_log_sum(
            *[
                eval
                for axis, eval
                in evals.items()
                if axis in axes
            ]
        )
    return log_err

def _get_trace_err(
    es: dict[Axis, np.ndarray],
    evals: dict[Axis, np.ndarray],
    all_axes: set[Axis]
) -> float:
    """
    Returns the trace portion of the error
    """
    trace_err: float = sum(
                es[axis]
                @ evals[axis]
                for axis in all_axes
            )
    return trace_err

def _get_reg_err(
    evals: dict[Axis, np.ndarray],
    evecs: dict[Axis, np.ndarray],
    regularizing: bool,
    regularizer: Regularizer
) -> float:
    """
    Returns the regularization penalty
    portion of the error
    """
    if regularizing:
        reg_err: float = regularizer.loss(
            evals,
            evecs,
        )
    else:
        reg_err: float = 0
    return reg_err


"""
This file wraps the core functionality of the GmGM algorithm into a single function.
"""
from __future__ import annotations

# Import core functionality of GmGM
from .core.core import direct_svd, direct_left_eigenvectors
from .core.core import calculate_eigenvalues, calculate_eigenvectors
from .core.preprocessing import center, clr_prost, create_gram_matrices
from .core.presparse_methods import recompose_sparse_precisions

# Used for typing
from typing import Optional, Literal
from .typing import Axis
from .dataset import Dataset
from .extras.regularizers import Regularizer
try:
    from anndata import AnnData
except ImportError:
    AnnData = None
try:
    from mudata import MuData
except ImportError:
    MuData = None

# For warnings
import warnings

def GmGM(
    dataset: Dataset | AnnData,
    to_keep: float | int | dict[Axis, float | int],
    random_state: Optional[int] = None,
    batch_size: Optional[int] = None,
    verbose: bool = False,
    # `center` parameters
    centering_method: Optional[Literal["avg-overall", "clr-prost"]] = None,
    # `create_gram_matrices` parameters
    use_nonparanormal_skeptic: bool = False,
    # `calculate_eigenvectors` parameters
    n_comps: Optional[int] = None,
    # `calculate_eigenvalues` parameters
    max_small_steps: int = 5,
    max_line_search_steps: int = 20,
    lr_init: float = 1.0,
    max_iter: int = 1000,
    tol: float = 1e-3,
    regularizer: Optional[Regularizer] = None,
    force_posdef: bool = True,
    verbose_every: int = 100,
    always_regularize: bool = False,
    check_overstep_each_iter: bool = False,
    # `recompose_sparse_positions` parameters
    threshold_method: Literal["overall", "rowwise", "rowwise-col-weighted"] = "rowwise-col-weighted",
    # from_AnnData/MuData parameters
    use_highly_variable: bool = False,
    key_added: str = "gmgm",
    use_abs_of_graph: bool = True,
):
    """
    Performs GmGM on the given dataset.
    """
    # Convert AnnData/MuData to Dataset (if relevant)
    is_anndata: bool = AnnData is not None and isinstance(dataset, AnnData)
    is_mudata: bool = MuData is not None and isinstance(dataset, MuData)
    if is_anndata:
        _dataset = Dataset.from_AnnData(dataset, use_highly_variable=use_highly_variable)
    elif is_mudata:
        _dataset = Dataset.from_MuData(dataset, use_highly_variable=use_highly_variable)
    else:
        _dataset = dataset

    # Save the random state
    _dataset.random_state = random_state

    # Center dataset
    if verbose:
        print("Centering...")
    if centering_method == "clr-prost":
        if is_anndata or is_mudata:
            if 'log1p' in dataset.uns.keys():
                warnings.warn(
                    "Dataset was log1p-transformed; clr-prost expects raw compositional (such as count) data"
                )
        clr_prost(_dataset)
    elif centering_method == "avg-overall":
        center(_dataset)
    elif centering_method is None:
        pass
    else:
        raise ValueError(f"Invalid centering method: {centering_method}")
    
    # Calculate eigenvectors
    if verbose:
        print("Calculating eigenvectors...")
    # We can skip the gram matrix calculation by doing SVD directly
    # Bringing memory usage down from O(n^2) to O(n) if `n_comps`` is O(1)
    # In my experience it slows you down if we want all eigenvectors, so
    # we only do this if `n_comps`` is specified
    unimodal: bool = len(_dataset.dataset) == 1
    matrix_variate: bool = all([
        _dataset.dataset[key].ndim == 2
        for key in _dataset.dataset.keys()
    ])
    # If dataset is a single matrix, then we can do SVD on said matrix
    # to get the eigenvectors for both axes at once
    if unimodal and matrix_variate and n_comps is not None:
        if verbose:
            print("\tby calculating SVD...")
        direct_svd(
            _dataset,
            n_comps=n_comps,
            random_state=random_state,
        )
    # If dataset is multi-modal or tensor-variate, we can find the left
    # eigenvectors of the concatenation of the matricization of each modality
    # on a given axis
    elif n_comps is not None:
        if verbose:
            print("\tby calculating left eigenvectors of concatenated matricizations...")
        direct_left_eigenvectors(
            _dataset,
            n_comps=n_comps,
            random_state=random_state,
        )
    # If dataset is multi-modal or is tensor-variate, we need to calculate the gram matrices
    # An O(n^2) memory operation
    else:
        if verbose:
            print("\tby calculating gram matrices and then eigendecomposing...")
        # Create Gram matrices
        create_gram_matrices(
            _dataset,
            use_nonparanormal_skeptic=use_nonparanormal_skeptic,
            batch_size=batch_size
        )

        # Calculate eigenvectors
        calculate_eigenvectors(
            _dataset,
            n_comps=n_comps,
            random_state=random_state,
            verbose=verbose
        )

    # Calculate eigenvalues
    if verbose:
        print("Calculating eigenvalues...")
    calculate_eigenvalues(
        _dataset,
        max_small_steps=max_small_steps,
        max_line_search_steps=max_line_search_steps,
        lr_init=lr_init,
        max_iter=max_iter,
        tol=tol,
        regularizer=regularizer,
        force_posdef=force_posdef,
        verbose=verbose,
        verbose_every=verbose_every,
        always_regularize=always_regularize,
        check_overstep_each_iter=check_overstep_each_iter,
    )

    # Recompose sparse precisions
    if verbose:
        print("Recomposing sparse precisions...")
    recompose_sparse_precisions(
        _dataset,
        to_keep=to_keep,
        threshold_method=threshold_method,
        batch_size=batch_size
    )

    # If was AnnData/MuData, return it
    if AnnData is not None and isinstance(dataset, AnnData):
        if verbose:
            print("Converting back to AnnData...")
        return _dataset.to_AnnData(key_added=key_added, use_abs_of_graph=use_abs_of_graph)
    elif MuData is not None and isinstance(dataset, MuData):
        if verbose:
            print("Converting back to MuData...")
        return _dataset.to_MuData(key_added=key_added, use_abs_of_graph=use_abs_of_graph)
    
    if verbose:
        print("Done!")
    
    # Otherwise, return a Dataset object
    return _dataset


"""
This file contains the methods that are of the following form:
    input -> densifying operation -> sparsifying projection -> output

An example would be a method that takes our eigenvectors and eigenvalues,
creates a precision matrix, and then thresholds that precision matrix.

Rather than ever creating the dense matrix, we can do the operations simultaneously,
preserving sparsity.
"""

from __future__ import annotations
from typing import Literal, Optional
import warnings

from ..typing import DataTensor
from ..dataset import Dataset, Axis
import scipy.sparse as sparse
import scipy.stats as stats
import numpy as np
import numba as nb

# ------------------------------------ #
#            Helper Methods            #
# ------------------------------------ #
@nb.jit(nopython=True, fastmath=True, parallel=True, cache=True)
def k_largest_per_row(
    matrix: np.ndarray,
    k: int
) -> np.ndarray:
    """
    Equivalent to
    np.argpartition(
        matrix,
        -k,
        axis=0
    )[-k:]

    But it is much faster. (At least on large matrices)
    """
    output = np.zeros((matrix.shape[0], k), dtype=np.int32)
    for i in nb.prange(matrix.shape[0]):
        row = matrix[i]
        
        to_fill = output[i]
        to_fill[:] = np.arange(k)
        values = row[:k]

        for j in range(k, row.size):
            if row[j] > values[0]:
                to_fill[0] = j
                values[0] = row[j]
                for l in range(1, k):
                    if values[l] > values[l-1]:
                        break
                    else:
                        to_fill[l-1], to_fill[l] = to_fill[l], to_fill[l-1]
                        values[l-1], values[l] = values[l], values[l-1]

    return output



# ------------------------------------ #
#              Public API              #
# ------------------------------------ #
def sparse_grammify(
    X: Dataset,
    to_keep: float | dict[Axis, float],
    threshold_method: Literal["overall", "rowwise", "rowwise-col-weighted"] = "overall",
    gram_method: Literal["covariance", "nonparanormal skeptic"] = "covariance",
    batch_size: int = 100,
    verbose: bool = False
) -> Dataset:
    """
    Creates a pre-thresholded gram matrix,
    meant for use with covariance subdivisions

    Modifies the dataset in-place
    """
    # Clear old gram matrices
    X.gram_matrices = {}

    # Calculate effective gram matrices
    num_samples: dict[Axis, int] = {}
    
    for modality, tensor in X.dataset.items():
        for idx, axis in enumerate(X.structure[modality]):
            if axis in X.batch_axes:
                continue

            matricized: np.ndarray | sparse.sparray
            if isinstance(tensor, sparse.sparray):
                # Handle sparse matrices - note that
                # these are always a matrix, and hence
                # have two axes and no batches.
                # In this case, one axis needs just the tensor
                # and the other needs the transpose of the tensor
                matricized: sparse.sparray = tensor
                if X.presences[axis][modality] == 1:
                    matricized = matricized.T
                matricized = matricized.tocsr()
            else:
                matricized: np.ndarray = np.reshape(
                    np.moveaxis(tensor, idx, 0),
                    (tensor.shape[idx], -1),
                    order='F'
                )

            if verbose:
                print(axis, modality)

            if gram_method == "covariance":
                if threshold_method == "overall":
                    to_add = _estimate_sparse_gram(
                        matricized,
                        to_keep[axis],
                        batch_size,
                        verbose=verbose
                    )
                elif threshold_method == "rowwise":
                    to_add = _estimate_sparse_gram_per_row(
                        matricized,
                        to_keep[axis],
                        batch_size,
                        verbose=verbose
                    )
                elif threshold_method == "rowwise-col-weighted":
                    to_add = _estimate_sparse_gram_per_row_col_weighted(
                        matricized,
                        to_keep[axis],
                        batch_size,
                        verbose=verbose
                    )
            elif gram_method == "nonparanormal skeptic":
                if threshold_method == "overall":
                    to_add = _estimate_sparse_nonparanormal_skeptic(
                        matricized,
                        to_keep[axis],
                        batch_size,
                        verbose=verbose
                    )
                else:
                    raise NotImplementedError("Nonparanormal Skeptic only supports overall thresholding")
            else:
                raise ValueError(
                    f"Unknown method {self.gram_method}"
                    + "\nPlease choose either 'covariance' or 'nonparanormal skeptic'"
                )
            new_samples = X.full_sizes[modality] // X.axis_sizes[axis]
            if axis in X.gram_matrices:
                X.gram_matrices[axis] += to_add
                num_samples[axis] += new_samples
            else:
                X.gram_matrices[axis] = to_add
                num_samples[axis] = new_samples

    # Divide gram matrices by effective number of modalities of samples seen
    for axis in X.gram_matrices.keys():
        X.gram_matrices[axis] /= num_samples[axis]

    # TODO: Remove this; but while I depend on igraph I cannot use sparse arrays
    # but rather I need sparse matrices
    X.gram_matrices = {
        axis: sparse.csr_array(mat)
        for axis, mat in X.gram_matrices.items()
    }

    return X

def recompose_sparse_precisions(
    X: Dataset,
    to_keep: float | int | dict[Axis, float | int],
    threshold_method: Literal["overall", "rowwise", "rowwise-col-weighted"] = "overall",
    batch_size: Optional[int] = None,
    verbose: bool = False
) -> Dataset:
    """
    Creates a pre-thresholded precision matrix
    """

    if isinstance(to_keep, float) | isinstance(to_keep, int):
        to_keep = {
            axis: to_keep
            for axis in X.all_axes
        }

    # Calculate the precision matrices
    X.precision_matrices: dict[Axis, np.ndarray] = {}
    for axis in X.all_axes:
        if verbose:
            print(f"Recomposing {axis}")

        # Ensure no negative eigenvalues,
        # necessary to allow square rooting
        negatives = X.evals[axis] < 0
        if negatives.any():
            warnings.warn(
                f"Found negative eigenvalues in axis {axis}"
                + "\nSetting them to 0"
            )
        X.evals[axis][negatives] = 0

        # half @ half.T will equal our precision matrix
        half = X.evecs[axis] * np.sqrt(X.evals[axis])

        # Simultaneously compute and threshold our precision matrix
        if threshold_method == "overall":
            X.precision_matrices[axis] = _estimate_sparse_gram(
                half,
                to_keep[axis],
                batch_size,
                verbose=verbose
            )
        elif threshold_method == "rowwise":
            X.precision_matrices[axis] = _estimate_sparse_gram_per_row(
                half,
                to_keep[axis],
                batch_size,
                verbose=verbose
            )
        elif threshold_method == "rowwise-col-weighted":
            X.precision_matrices[axis] = _estimate_sparse_gram_per_row_col_weighted(
                half,
                to_keep[axis],
                batch_size,
                verbose=verbose
            )
        else:
            raise ValueError(f"{threshold_method} is not a valid thresholding method")
        
        # Symmetricize precision matrix
        X.precision_matrices[axis] = (
            X.precision_matrices[axis]
            + X.precision_matrices[axis].T
        ) / 2
        
    return X

# ------------------------------------ #
#            Helper Methods            #
# ------------------------------------ #
def _estimate_sparse_gram(
    matrix: DataTensor,
    percent_to_keep: float,
    batch_size: Optional[int] = None,
    divide_by_diagonal: bool = False,
    verbose: bool = False
) -> sparse.sparray:
    """
    Computes matrix @ matrix.T in a sparse manner.
    """

    if batch_size is None:
        batch_size = min(matrix.shape[0], 1000)

    assert percent_to_keep >= 0 and percent_to_keep <= 1, \
        "`percent_to_keep` must be between 0 and 1"

    features, _ = matrix.shape
    num_to_keep: int = int(percent_to_keep * (features**2 - features) / 2)

    # Create a sparse array to store the result
    # We are always gonna keep it so smallest kept value is at `num_to_keep`
    # Note that this uses twice as much memory as we ultimately will need,
    # because we want to zipper-sort the old data and new data in the end.
    # If there is a vectorized way to zipper-sort, rather than concatenating them,
    # then we can save this memory.  But idk how!
    result = sparse.coo_array(
        (
            np.zeros(2*num_to_keep, dtype=np.float32),
            (
                np.zeros(2*num_to_keep, dtype=np.int32),
                np.zeros(2*num_to_keep, dtype=np.int32)
            )
        ),
        shape=(features, features),
    )


    for i in range(0, features + batch_size, batch_size):
        if i >= features:
            break
        # Compute the entire row at once to save time
        # Assuming we have space that is linear in axis length
        # (a reasonable assumption)
        # then this is okay to do!
        increase: int = min(batch_size, features - i)
        res_batch: np.ndarray
        if isinstance(matrix, sparse.sparray):
            # Warning: this implementation becomes quite slow
            # when the matrices are very large.  Better to,
            # if possible, work with dense matrices.
            rhs = matrix[i:i+increase].toarray().T
            res_batch = np.abs(matrix[i:] @ rhs)
        else:
            res_batch = np.abs(matrix[i:] @ matrix[i:i+increase].T)

        # Remove the diagonal so it does not affect thresholding!
        diags = np.diagonal(res_batch).copy()
        if divide_by_diagonal:
            diags[diags == 0] = 1
            np.fill_diagonal(res_batch, 0)
        else:
            diags[:] = 1

        # Remove everything above diagonal so it does not affect thresholding!
        # (due to symmetry)
        res_batch[np.triu_indices(increase)] = 0

        # We only need to check the `num_to_keep` largest elements
        # in this batch!
        to_check = min(num_to_keep, res_batch.size)
        flat_res_batch = res_batch.reshape(-1)
        largest_idxs = np.argpartition(
            (res_batch / diags).reshape(-1),
            -to_check,
            axis=None
        )[-to_check:]
        rows = largest_idxs // res_batch.shape[1]
        cols = largest_idxs % res_batch.shape[1]

        # Remove any elements that are smaller than the smallest kept
        candidates = flat_res_batch[largest_idxs] > result.data[num_to_keep]
        flat_res_batch = flat_res_batch[largest_idxs][candidates]
        rows = rows[candidates]
        cols = cols[candidates]

        if flat_res_batch.size == 0:
            # Nothing to add!
            continue

        # Figure out how many elements we can add
        sorted = np.argsort(flat_res_batch)
        flat_res_batch = flat_res_batch[sorted]
        rows = rows[sorted]
        cols = cols[sorted]
        amount = (result.data < flat_res_batch[-1]).sum()
        amount = min(flat_res_batch.size, amount)

        if amount == 0:
            # Nothing to add!
            continue

        # Add these to the data!
        result.data[:amount] = flat_res_batch[-amount:]
        result.row[:amount] = rows[-amount:] + i
        result.col[:amount] = cols[-amount:] + i

        # And sort the data
        # Note that this makes the complexity klogk where k is the number
        # of elements to keep; this is slower than the older implementation
        # using dictionaries and dok_arrays.
        # However, because everything is vectorized here, and k is typically
        # quite small, and log k is smaller still, this is MUCH faster in practice.
        sorted = np.argsort(result.data)
        result.data = result.data[sorted]
        result.row = result.row[sorted]
        result.col = result.col[sorted]

        # Bottleneck (>75-80% runtime) for large (10,000x10,000) data is the matrix multiplication
        # when keeping on average 10 edges per row (i.e. k=100,000).
        # The sorts account for ~5-10% of runtime.
        #
        # Argpartition is linear time in the number of elements to partition, and it takes ~15%.

    # Remove the excess memory
    result.data = result.data[num_to_keep:]
    result.row = result.row[num_to_keep:]
    result.col = result.col[num_to_keep:]

    return result

def _estimate_sparse_gram_per_row(
    matrix: DataTensor,
    per_row_to_keep: float,
    batch_size: Optional[int] = None,
    verbose: bool = False
) -> sparse.sparray:
    """
    Computes matrix @ matrix.T in a sparse manner.
    """

    if batch_size is None:
        batch_size = min(matrix.shape[0], 1000)

    assert per_row_to_keep >= 0 and per_row_to_keep <= matrix.shape[0], \
        "`per_row_to_keep` must be between 0 and the number of features"

    features, _ = matrix.shape
    num_to_keep: int = per_row_to_keep * features

    # Create a sparse array to store the result
    result = sparse.coo_array(
        (
            np.zeros(num_to_keep, dtype=np.float32),
            (
                np.zeros(num_to_keep, dtype=np.int32),
                np.zeros(num_to_keep, dtype=np.int32)
            )
        ),
        shape=(features, features),
    )


    for i in range(0, features + batch_size, batch_size):
        if i >= features:
            break
        # Compute the entire row at once to save time
        # Assuming we have space that is linear in axis length
        # (a reasonable assumption)
        # then this is okay to do!
        increase: int = min(batch_size, features - i)
        res_batch: np.ndarray
        if isinstance(matrix, sparse.sparray):
            # Warning: this implementation becomes quite slow
            # when the matrices are very large.  Better to,
            # if possible, work with dense matrices.
            rhs = matrix[i:i+increase].toarray().T
            res_batch = np.abs(matrix @ rhs)
        else:
            res_batch = np.abs(matrix @ matrix[i:i+increase].T)

        # Remove the diagonal so it does not affect selection!
        np.fill_diagonal(res_batch[i:], 0)

        # We only need to check the `num_to_keep` largest elements
        # in this batch!
        largest_idxs = k_largest_per_row(
            res_batch.T,
            per_row_to_keep,
        ).T
        # largest_idxs = np.argpartition(
        #     res_batch,
        #     -per_row_to_keep,
        #     axis=0
        # )[-per_row_to_keep:]
        rows = largest_idxs.T.reshape(-1)

        # Add these to the data!
        start = i * per_row_to_keep
        end = (i + increase) * per_row_to_keep
        result.data[start:end] = np.take_along_axis(
            res_batch,
            largest_idxs,
            axis=0
        ).T.reshape(-1)
        result.row[start:end] = rows

        for b in range(increase):
            start = (i + b) * per_row_to_keep
            end = (i + 1 + b) * per_row_to_keep
            result.col[start:end] = i + b

    # Remove upper triangle
    result.data[result.row < result.col] = 0

    return result

def _estimate_sparse_gram_per_row_col_weighted(
    matrix: DataTensor,
    per_row_to_keep: float,
    batch_size: Optional[int] = None,
    verbose: bool = False
) -> sparse.sparray:
    """
    Computes matrix @ matrix.T in a sparse manner.
    """

    if batch_size is None:
        batch_size = min(matrix.shape[0], 1000)

    assert per_row_to_keep >= 0 and per_row_to_keep <= matrix.shape[0], \
        "`per_row_to_keep` must be between 0 and the number of features"

    features, _ = matrix.shape
    num_to_keep: int = per_row_to_keep * features

    # Construct column weights - because we are using absolute value,
    # this is more complicated!
    # Equivalent to:
    #   col_weights = np.abs(matrix @ matrix.T).sum(axis=0)
    col_weights = np.zeros(features, dtype=np.float32)
    for i in range(0, features + batch_size, batch_size):
        if i >= features:
            break
        # Compute the entire row at once to save time
        # Assuming we have space that is linear in axis length
        # (a reasonable assumption)
        # then this is okay to do!
        increase: int = min(batch_size, features - i)
        res_batch: np.ndarray
        if isinstance(matrix, sparse.sparray):
            # Warning: this implementation becomes quite slow
            # when the matrices are very large.  Better to,
            # if possible, work with dense matrices.
            rhs = matrix[i:i+increase].toarray().T
            res_batch = np.abs(matrix @ rhs)
        else:
            res_batch = np.abs(matrix @ matrix[i:i+increase].T)

        # Remove the diagonal so it does not affect the sum!
        np.fill_diagonal(res_batch[i:], 0)

        col_weights[i:i+increase] += res_batch.sum(axis=0)


    col_weights[col_weights == 0] = 1
    col_weights = col_weights.reshape(-1, 1)

    # Create a sparse array to store the result
    result = sparse.coo_array(
        (
            np.zeros(num_to_keep, dtype=np.float32),
            (
                np.zeros(num_to_keep, dtype=np.int32),
                np.zeros(num_to_keep, dtype=np.int32)
            )
        ),
        shape=(features, features),
    )


    for i in range(0, features + batch_size, batch_size):
        if i >= features:
            break
        # Compute the entire row at once to save time
        # Assuming we have space that is linear in axis length
        # (a reasonable assumption)
        # then this is okay to do!
        increase: int = min(batch_size, features - i)
        res_batch: np.ndarray
        if isinstance(matrix, sparse.sparray):
            # Warning: this implementation becomes quite slow
            # when the matrices are very large.  Better to,
            # if possible, work with dense matrices.
            rhs = matrix[i:i+increase].toarray().T
            res_batch = np.abs(matrix @ rhs)
        else:
            res_batch = np.abs(matrix @ matrix[i:i+increase].T)

        # Remove the diagonal so it does not affect selection!
        np.fill_diagonal(res_batch[i:], 0)

        # We only need to check the `per_row_to_keep` largest elements
        # in this batch!
        largest_idxs = k_largest_per_row(
            (res_batch / col_weights).T,
            per_row_to_keep,
        ).T
        rows = largest_idxs.T.reshape(-1)

        # print(res_batch.shape)
        # print(per_row_to_keep)
        # print((res_batch / col_weights)[largest_idxs])
        # print((res_batch / col_weights).max(axis=0))
        # print("---")

        # Add these to the data!
        start = i * per_row_to_keep
        end = (i + increase) * per_row_to_keep
        result.data[start:end] = np.take_along_axis(
            res_batch,
            largest_idxs,
            axis=0
        ).T.reshape(-1)
        result.row[start:end] = rows

        for b in range(increase):
            start = (i + b) * per_row_to_keep
            end = (i + 1 + b) * per_row_to_keep
            result.col[start:end] = i + b

    # Remove upper triangle
    #result.data[result.row < result.col] = 0

    # #print(result.data)
    # print(result.row)
    # print(result.col)

    return result

def _estimate_sparse_nonparanormal_skeptic(
    matrix: DataTensor,
    percent_to_keep: float,
    batch_size: Optional[int] = None,
    divide_by_diagonal: bool = True,
    verbose: bool = False
) -> sparse.sparray:
    """
    Computes matrix @ matrix.T in a sparse manner.
    """

    if batch_size is None:
        batch_size = min(matrix.shape[0], 1000)

    assert percent_to_keep >= 0 and percent_to_keep <= 1, \
        "`percent_to_keep` must be between 0 and 1"

    features, _ = matrix.shape
    num_to_keep: int = int(percent_to_keep * (features**2 - features) / 2)

    # Create a sparse array to store the result
    # We are always gonna keep it so smallest kept value is at `num_to_keep`
    # Note that this uses twice as much memory as we ultimately will need,
    # because we want to zipper-sort the old data and new data in the end.
    # If there is a vectorized way to zipper-sort, rather than concatenating them,
    # then we can save this memory.  But idk how!
    result = sparse.coo_array(
        (
            np.zeros(2*num_to_keep, dtype=np.float32),
            (
                np.zeros(2*num_to_keep, dtype=np.int32),
                np.zeros(2*num_to_keep, dtype=np.int32)
            )
        ),
        shape=(features, features),
    )

    # Replace matrix with ranks
    matrix = stats.rankdata(matrix, axis=1, nan_policy='raise')
    matrix = matrix - matrix.mean(axis=1, keepdims=True)

    # Get square root of diagonals of matrix
    sqrt_diags = np.zeros(features, dtype=np.float32)
    for i in range(0, features):
        sqrt_diags[i] = 1/np.sqrt(matrix[i] @ matrix[i])

    for i in range(0, features + batch_size, batch_size):
        if i >= features:
            break
        # Compute the entire row at once to save time
        # Assuming we have space that is linear in axis length
        # (a reasonable assumption)
        # then this is okay to do!
        increase: int = min(batch_size, features - i)
        res_batch: np.ndarray

        # Calculate correlation coefficient of matrix of ranks
        res_batch = matrix[i:] @ matrix[i:i+increase].T
        res_batch = (
            sqrt_diags[i:i+increase].reshape(1, -1)
            * res_batch
            * sqrt_diags[i:].reshape(-1, 1)
        )
        res_batch = np.abs(2 * np.sin(np.pi / 6 * res_batch))

        # Remove the diagonal so it does not affect thresholding!
        np.fill_diagonal(res_batch, 0)

        # Remove everything above diagonal so it does not affect thresholding!
        # (due to symmetry)
        res_batch[np.triu_indices(increase)] = 0

        # We only need to check the `num_to_keep` largest elements
        # in this batch!
        to_check = min(num_to_keep, res_batch.size)
        flat_res_batch = res_batch.reshape(-1)
        largest_idxs = np.argpartition(
            flat_res_batch,
            -to_check,
            axis=None
        )[-to_check:]
        rows = largest_idxs // res_batch.shape[1]
        cols = largest_idxs % res_batch.shape[1]

        # Remove any elements that are smaller than the smallest kept
        candidates = flat_res_batch[largest_idxs] > result.data[num_to_keep]
        flat_res_batch = flat_res_batch[largest_idxs][candidates]
        rows = rows[candidates]
        cols = cols[candidates]

        if flat_res_batch.size == 0:
            # Nothing to add!
            continue

        # Figure out how many elements we can add
        sorted = np.argsort(flat_res_batch)
        flat_res_batch = flat_res_batch[sorted]
        rows = rows[sorted]
        cols = cols[sorted]
        amount = (result.data < flat_res_batch[-1]).sum()
        amount = min(flat_res_batch.size, amount)

        if amount == 0:
            # Nothing to add!
            continue

        # Add these to the data!
        result.data[:amount] = flat_res_batch[-amount:]
        result.row[:amount] = rows[-amount:] + i
        result.col[:amount] = cols[-amount:] + i

        # And sort the data
        # Note that this makes the complexity klogk where k is the number
        # of elements to keep; this is slower than the older implementation
        # using dictionaries and dok_arrays.
        # However, because everything is vectorized here, and k is typically
        # quite small, and log k is smaller still, this is MUCH faster in practice.
        sorted = np.argsort(result.data)
        result.data = result.data[sorted]
        result.row = result.row[sorted]
        result.col = result.col[sorted]

        # Bottleneck (>75-80% runtime) for large (10,000x10,000) data is the matrix multiplication
        # when keeping on average 10 edges per row (i.e. k=100,000).
        # The sorts account for ~5-10% of runtime.
        #
        # Argpartition is linear time in the number of elements to partition, and it takes ~15%.

    # Remove the excess memory
    result.data = result.data[num_to_keep:]
    result.row = result.row[num_to_keep:]
    result.col = result.col[num_to_keep:]

    return result